package cn.fw.hermes.service.biz;

import cn.fw.common.cache.locker.DistributedLocker;
import cn.fw.hermes.common.constant.ActionStatusEnum;
import cn.fw.hermes.common.utils.StringUtils;
import cn.fw.hermes.common.utils.UUIDUtils;
import cn.fw.hermes.common.utils.im.RedisKeyUtil;
import cn.fw.hermes.common.utils.im.TXCloudHelper;
import cn.fw.hermes.common.utils.im.TXCloudUtils;
import cn.fw.hermes.domain.db.Account;
import cn.fw.hermes.domain.dto.AdminConfigDto;
import cn.fw.hermes.domain.dto.IdentifierDto;
import cn.fw.hermes.domain.enums.AccountTypeEnum;
import cn.fw.hermes.domain.enums.UserTypeEnum;
import cn.fw.hermes.domain.vo.AccountResultVO;
import cn.fw.hermes.sdk.api.para.AccountCondition;
import cn.fw.hermes.sdk.api.result.AccountResult;
import cn.fw.hermes.service.data.AccountService;
import com.alibaba.fastjson.JSONObject;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.tuple.Pair;
import org.redisson.api.RLock;
import org.springframework.beans.BeanUtils;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.lang.NonNull;
import org.springframework.stereotype.Service;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import static cn.fw.common.businessvalidator.Validator.BV;


/**
 * @Author: Chenery
 * @Date: 2021/1/27 16:21
 */
@RequiredArgsConstructor
@Slf4j
@Service
public class AccountBizService {

    private final DistributedLocker distributedLocker;
    private final StringRedisTemplate redisTemplate;
    private final AccountService accountService;
    private final AdminConfigDto adminConfigDto;

    /**
     * 根据用户类型注册
     *
     * @param accountCondition 注册账户对象
     * @return 注册结果
     */
    public AccountResult userRegister(@NonNull AccountCondition accountCondition) {
        Account account = new Account();
        BeanUtils.copyProperties(accountCondition, account);
        account.setStatus(true);
        account.setAccountType(AccountTypeEnum.ORDINARY);
        account.setNickName(accountCondition.getUserName());
        account.setUserType(accountCondition.isStaff() ? UserTypeEnum.B : UserTypeEnum.C);
        Optional<Account> optional = accountService
                .queryAccountByUserId(account.getUserId(), account.getUserType().getValue());
        //判断用户是否已经注册 ，如已注册，直接返回当前已注册账户
        if (optional.isPresent()) {
            AccountResult accountResult = new AccountResult();
            BeanUtils.copyProperties(optional.get(), accountResult);
            return accountResult;
        }
        String identifier =
                accountCondition.isStaff() ? TXCloudHelper.getIdentifier(account.getUserId(),
                        TXCloudHelper.PrefixEnum.BUSINESS)
                        : TXCloudHelper.getIdentifier(account.getUserId(),
                                TXCloudHelper.PrefixEnum.CUSTOMER);
        account.setIdentifierCode(identifier);
        //调用真正的注册实现方法
        return this.register(account);
    }

    /**
     * 通用用户注册
     *
     * @param account 账户对象
     * @return 注册后的账户对象
     */
    private AccountResult register(@NonNull Account account) {
        Integer expire = 3600 * 24 * 10;
        Pair<Boolean, RLock> lockPair = distributedLocker
                .tryLock(RedisKeyUtil.getDistributedKey(account.getIdentifierCode()),
                        TimeUnit.SECONDS, 0, 2);
        BV.isTrue(lockPair.getKey(), "userId：" + account.getUserId() + "的用户正在注册，请稍后重试");
        String adminKey = RedisKeyUtil.getAdminkey(adminConfigDto.getAdminId());
        //获取管理员签名
        String adminSig = getSig(adminConfigDto.getAdminId(), expire, adminKey).getSig();
        JSONObject jsonObject = TXCloudUtils.AccountImport(account.getIdentifierCode(),
                account.getNickName(), account.getFaceUrl(), adminSig, adminConfigDto.getAdminId(),
                adminConfigDto.getAppId());
        //是否注册成功
        BV.isTrue(jsonObject.get("ActionStatus").equals(ActionStatusEnum.OK.getCode()), "调用IM接口失败");
        //将注册成功的账户信息对象持久化到数据库
        boolean result = accountService.save(account);
        if (!result) {
            log.error("保存账户{}到数据库失败", account);
        }
        AccountResult accountResult = new AccountResult();
        BeanUtils.copyProperties(account, accountResult);
        return accountResult;

    }

    /**
     * 获取用户签名实现
     *
     * @param identifier 用户识别码
     * @param expire     签名过期时间
     * @param userKey    rediskey
     * @return 签名对象
     */
    public IdentifierDto getSig(@NonNull String identifier, @NonNull Integer expire,
            @NonNull String userKey) {
        IdentifierDto identifierDto = new IdentifierDto();
        identifierDto.setIdentifierCode(identifier);
        //先从redis中拿该用户签名
        String sig = redisTemplate.opsForValue().get(userKey);
        identifierDto.setSig(sig);
        //若签名信息不存在则获取签名，若存在直接返回
        if (StringUtils.isEmpty(sig)) {
            Pair<Boolean, RLock> lock = distributedLocker
                    .tryLock(RedisKeyUtil.getDistributedKey(identifier), TimeUnit.SECONDS, 0, 2);
            BV.isTrue(lock.getKey(), "identifier： " + identifier + "的用户正在获取签名");
            //将获取签名用户踢线下
            if (!adminConfigDto.getAdminId().equals(identifier)) {
                this.kick(identifier);
            }
            //调用腾讯IM接口获取签名
            sig = TXCloudHelper
                    .getSig(identifier, expire, adminConfigDto.getAppId(),
                            adminConfigDto.getPrivStr());
            BV.isTrue(!StringUtils.isEmpty(sig), "签名获取失败");
            redisTemplate.opsForValue().set(userKey, sig, expire, TimeUnit.SECONDS);
            identifierDto.setSig(sig);
        }
        return identifierDto;
    }


    /**
     * 获取非管理员用户签名
     *
     * @param userId 用户ID
     * @return 签名
     */
    public AccountResultVO getLoginSig(Long userId, Boolean isStaff) {
        String identifierCode;
        Optional<Account> optional = accountService.queryAccountByUserId(userId,
                isStaff ? UserTypeEnum.B.getValue() : UserTypeEnum.C.getValue());
        //判断当前获取签名用户是否注册 如果未注册将该用户注册
        if (!optional.isPresent()) {
            AccountCondition account = new AccountCondition();
            account.setUserId(userId);
            account.setStaff(isStaff);
            //当获取签名的用户未注册时用户昵称组成为：userId+uuid前6位
            account.setUserName(userId.toString() + UUIDUtils.newRandomUUid().substring(0, 6));
            identifierCode = this.userRegister(account).getIdentifierCode();
        } else {
            identifierCode = optional.get().getIdentifierCode();
        }
        IdentifierDto identifierDto = this
                .getSig(identifierCode, 3600 * 24 * 10, RedisKeyUtil.getUserkey(identifierCode));
        AccountResultVO vo = new AccountResultVO();
        vo.setUserId(userId);
        vo.setUserSig(identifierDto.getSig());
        vo.setIdentifier(identifierCode);
        return vo;
    }

    public String getSig(Long userId, Boolean isStaff) {
        return getLoginSig(userId, isStaff).getUserSig();
    }

    /**
     * 将用户踢下线 默认将用户类型为业务人员的用户踢下线
     *
     * @param userId 用户ID
     * @return 踢下线是否成功
     */
    public Boolean kick(@NonNull Long userId) {
        Optional<Account> optional = accountService
                .queryAccountByUserId(userId, UserTypeEnum.B.getValue());
        return optional.map(account -> this.kick(account.getIdentifierCode())).orElse(false);
    }

    /**
     * 踢用户下线
     *
     * @param identifier 用户识别码
     * @return 踢下线结果
     */
    private Boolean kick(@NonNull String identifier) {
        BV.isTrue(identifier.length() < 32, "用户识别码长度过长");
        //被踢用户是已否注册
        boolean exist = accountService.isExist(identifier);
        if (exist) {
            String adminKey = RedisKeyUtil.getAdminkey(adminConfigDto.getAdminId());
            String adminSig = getSig(adminConfigDto.getAdminId(), 3600 * 24 * 30, adminKey)
                    .getSig();
            JSONObject jsonObject = TXCloudUtils
                    .Kick(identifier, adminSig, adminConfigDto.getAdminId(),
                            adminConfigDto.getAppId());
            return jsonObject.get("ActionStatus").equals(ActionStatusEnum.OK.getCode());
        }
        return false;
    }

    /**
     * 通过userId查询用户identifierCode
     *
     * @param userId  用户Id
     * @param isStaff 是否为业务人员
     * @return
     */
    public String queryIdentifierCode(Long userId, Boolean isStaff) {
        BV.isTrue(!StringUtils.isEmpty(userId), "userId不能为空");
        Optional<Account> optional = accountService.queryAccountByUserId(userId,
                isStaff ? UserTypeEnum.B.getValue() : UserTypeEnum.C.getValue());
        BV.isTrue(optional.isPresent() && !StringUtils.isEmpty(optional.get().getIdentifierCode()),
                "用户不存在");
        return optional.get().getIdentifierCode();

    }


}
